#
# $Id$
#
# Test SMTP class.
#
# This class sends test emails to an MTA.  It also sends RSET, and
# reopens the SMTP connection at various points in sending the
# messages.
#
# The configuration is read from a file specified on the constructor.
# The example configuration file sendTestMessages.conf documents each
# configuration parameter.
#
# Author: Alban Deniz
#
# Copyright (C) 2009 by MicroNeil Corporation. All rights reserved.
# See www.armresearch.com for the copyright terms.
#
####################################################################

package TestSMTP;
use strict;
use IO::File;
use Net::SMTP;

sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    $self->{confFile} = shift;
    bless $self, $class;

    # Create defaults.
    $self->{hostname} = "localhost";
    $self->{message_directory} = "testMessages";
    $self->{message} = [];
    $self->{n_sets_send} = 1;
    $self->{use_pipelining} = "no";
    $self->{from_address} = [];
    $self->{to_address} = [];
    $self->{send_RSET_after_RCPT} = 0;
    $self->{send_RSET_after_MAIL} = 0;
    $self->{reopen_connection_after_MAIL} = 0;
    $self->{reopen_connection_after_RCPT} = 0;
    $self->{reopen_connection_after_DATA} = 0;
    $self->initialize();
    return $self;
}

# Open and read the configuration file.
sub initialize {
    my $self = shift;
    print "Using config file:  $self->{confFile}\n";
    my $fh = IO::File->new($self->{confFile}, "r") ||
	die "Unable to open $self->{confFile}: $!";
    my $line = "";
    my $key = "";
    my $lineNum = 0;

    while (1 == 1) {
	$line = $fh->getline();

	# Exit loop if there is no line.
	if (!defined($line)) {
	    last;
	}
	$lineNum++;

	# Process the line.
	$line =~ s/^\s+//;
	$line =~ s/\s+$//;
	my @field = split(/\s+/, $line);

	# Check for blank line.
	if (0 == @field) {
	    next;
	}
	$key = $field[0];

	# Check for comment.
	$DB::single=2;
	if (substr($key, 0, 1) eq "#") {
	    next;
	}

	# Neither blank line nor comment.  This is a valid
	# configuration line.
	(2 == @field) || die "Error on line $lineNum in file ".
	    "$self->{confFile}:\n\n$line\n\nEach line must consist ".
	    "of a keyword followed by one value (separated by white space)";

	my @keywords = qw(hostname message_directory n_sets_send
                          use_pipelining
                          send_RSET_after_RCPT send_RSET_after_MAIL
                          reopen_connection_after_MAIL
                          reopen_connection_after_RCPT
                          reopen_connection_after_DATA);

	foreach my $k (@keywords) {
	    if ($k eq $key) {
		$self->{$k} = $field[1];
		next;
	    }
	}
	if ("from_address" eq $key) {
	    $self->{from_address} = [@{$self->{from_address}}, $field[1]];
	} elsif ("to_address" eq $key) {
	    $self->{to_address} = [@{$self->{to_address}}, $field[1]];
	} else {
	    "Error on line $lineNum in file ".
		"$self->{confFile}:\n\n$line\n\nThe keyword ".
		"(first word on the line) is invalid";
	}

    };
    $fh->close() || die "Unable to close $self->{confFile}: $!";

    # Read the messages.
    $self->readMessages();
}

# Display the configuration parameters.
sub displayParameters {
    my $self = shift;
    print "Configuration--\n";
    print "\thostname:  $self->{hostname}\n";
    print "\tmessage_directory:  $self->{message_directory}\n";
    print "\tn_sets_send:  $self->{n_sets_send}\n";
    print "\tuse_pipelining:  $self->{use_pipelining} (NOT IMPLEMENTED)\n";
    print "\tsend_RSET_after_RCPT:  $self->{send_RSET_after_RCPT}\n";
    print "\tsend_RSET_after_MAIL:  $self->{send_RSET_after_MAIL}\n";
    print "\treopen_connection_after_MAIL:  " .
	$self->{reopen_connection_after_MAIL} ."\n";
    print "\treopen_connection_after_RCPT:  " .
	$self->{reopen_connection_after_RCPT} . "\n";
    print "\treopen_connection_after_DATA:  " .
	$self->{reopen_connection_after_DATA} ."\n";
    print "\tTo: addresses--\n";
    foreach my $to (@{$self->{to_address}}) {
	print "\t\t$to\n";
    }
    print "\tFrom: addresses--\n";
    foreach my $from (@{$self->{from_address}}) {
	print "\t\t$from\n";
    }
    print "Read " . $self->nMessages() . " messages.\n";
}

# Read the email messages.
sub readMessages {
    my $self = shift;

    $self->{message} = [];

    opendir(msgDir, $self->{message_directory}) ||
	die "Unable to open $self->{message_directory}: $!";

    foreach my $fileName (sort readdir(msgDir)) {

	if (substr($fileName, 0, 1) eq ".") {
	    next;
	}

	$fileName = $self->{message_directory} . "/" . $fileName;

	# Open the email message.
	my $fh = IO::File->new($fileName, "r") ||
	    die "Unable to open $fileName: $!";

	# Read it.
	$self->{message} = [@{$self->{message}}, join("", $fh->getlines())];

	# Close.
	$fh->close() || die "Unable to close $fileName: $!";

    }

    closedir(msgDir) ||
	die "Unable to close $self->{message_directory}: $!";

}

# Return the number of messages read.
sub nMessages {
    my $self = shift;
    return scalar @{$self->{message}};
}

# Begin sending messages.
sub begin {
    my $self = shift;

    # Index of next from_address to use.
    my $fromAddressIndex = 0;
    my $origFromAddressIndex;

    # Index of next to_address to use.
    my $toAddressIndex = 0;
    my $origToAddressIndex;

    my $smtp = Net::SMTP->new($self->{hostname}) ||
	die "Unable to establish SMTP connection with $self->{hostname}: $!";

    # Number of messages sent since the last SMTP connection close
    # after DATA.
    my $nMessagesSince_DATA_Close = 0;

    # Number of messages sent since the last SMTP connection close
    # after MAIL.
    my $nMessagesSince_MAIL_Close = 0;

    # Number of messages sent since the last SMTP connection close
    # after RCPT.
    my $nMessagesSince_RCPT_Close = 0;

    # Number of messages sent since the last SMTP RSET was sent after
    # MAIL.
    my $nMessagesSince_MAIL_RSET = 0;

    # Number of messages sent since the last SMTP RSET was sent after
    # RCPT.
    my $nMessagesSince_RCPT_RSET = 0;

    # Send messages.
    my $iSet = 0;
    while (1 == 1) {

	# Increment set index.
	$iSet++;

	# Exit?
	if ( ($self->{n_sets_send} != 0) &&
	     ($iSet > $self->{n_sets_send})) {
		last;
	}

	for (my $iMsg = 0; $iMsg < $self->nMessages(); $iMsg++) {

	    # Save indices for the from and to addresses.
	    $origFromAddressIndex = $fromAddressIndex;
	    $origToAddressIndex = $toAddressIndex;

	    # Load fromAddress.
	    my $fromAddress = ${$self->{from_address}}[$fromAddressIndex++];

	    # Reset fromAddress index.
	    if (@{$self->{from_address}} == $fromAddressIndex) {
		$fromAddressIndex = 0;
	    }

	    # Load toAddress.
	    my $toAddress = ${$self->{to_address}}[$toAddressIndex++];

	    # Reset toAddress index.
	    if (@{$self->{to_address}} == $toAddressIndex) {
		$toAddressIndex = 0;
	    }

	    my $msgNum = $iMsg + 1;
	    print "Sending message $msgNum from iteration $iSet " .
		"(to $toAddress, from $fromAddress)...";
	    $smtp->mail($fromAddress) ||
		die "Error from Net::SMTP::mail(): $!";

	    # Close/reopen SMTP connection after MAIL?
	    $nMessagesSince_MAIL_Close++;
	    if ($nMessagesSince_MAIL_Close ==
		$self->{reopen_connection_after_MAIL}) {

		print "Closing and reopening the SMTP connection after ".
		    "MAIL...";
		$nMessagesSince_MAIL_Close = 0;
		$smtp->quit() ||
		    die "Unable to close SMTP connection:  $!";
		
		$smtp = Net::SMTP->new($self->{hostname}) ||
		    die "Unable to establish SMTP connection with " .
		    "$self->{hostname}: $!";
		print "done.\n";

		# Restore indices for the from and to addresses.
		$fromAddressIndex = $origFromAddressIndex;
		$toAddressIndex = $origToAddressIndex;
		redo;
	    }

	    # Send RSET after MAIL?
	    $nMessagesSince_MAIL_RSET++;
	    if ($nMessagesSince_MAIL_RSET == $self->{send_RSET_after_MAIL}) {

		print "Sending RSET after MAIL...";
		$nMessagesSince_MAIL_RSET = 0;
		$smtp->reset() ||
		    die "Error sending RSET:  $!";
		print "done.\n";

		# Restore indices for the from and to addresses.
		$fromAddressIndex = $origFromAddressIndex;
		$toAddressIndex = $origToAddressIndex;
		redo;
	    }

	    $smtp->recipient($toAddress) ||
		die "Error from Net::SMTP::recipient(): $!";

	    # Close/reopen SMTP connection after RCPT?
	    $nMessagesSince_RCPT_Close++;
	    if ($nMessagesSince_RCPT_Close ==
		$self->{reopen_connection_after_RCPT}) {

		print "Closing and reopening the SMTP connection after ".
		    "RCPT...";
		$nMessagesSince_RCPT_Close = 0;
		$smtp->quit() ||
		    die "Unable to close SMTP connection:  $!";
		
		$smtp = Net::SMTP->new($self->{hostname}) ||
		    die "Unable to establish SMTP connection with " .
		    "$self->{hostname}: $!";
		print "done.\n";

		# Restore indices for the from and to addresses.
		$fromAddressIndex = $origFromAddressIndex;
		$toAddressIndex = $origToAddressIndex;
		redo;
	    }

	    # Send RSET after RCPT?
	    $nMessagesSince_RCPT_RSET++;
	    if ($nMessagesSince_RCPT_RSET == $self->{send_RSET_after_RCPT}) {

		print "Sending RSET after RCPT...";
		$nMessagesSince_RCPT_RSET = 0;
		$smtp->reset() ||
		    die "Error sending RSET:  $!";
		print "done.\n";

		# Restore indices for the from and to addresses.
		$fromAddressIndex = $origFromAddressIndex;
		$toAddressIndex = $origToAddressIndex;
		redo;
	    }

	    $smtp->data(${$self->{message}}[$iMsg]) ||
		die "Error from Net::SMTP::data(): $!";
	    print "done.\n";

	    # Close/reopen SMTP connection after DATA?
	    $nMessagesSince_DATA_Close++;
	    if ($nMessagesSince_DATA_Close ==
		$self->{reopen_connection_after_DATA}) {

		print "Closing and reopening the SMTP connection...";
		$nMessagesSince_DATA_Close = 0;
		$smtp->quit() ||
		    die "Unable to close SMTP connection:  $!";
		
		$smtp = Net::SMTP->new($self->{hostname}) ||
		    die "Unable to establish SMTP connection with " .
		    "$self->{hostname}: $!";
		print "done.\n";
	    }
	    
	}
    }

    $smtp->quit();
}
1;
