// $Id$
//
// \file
//
// Copyright (C) 2008, ARM Research Labs, LLC.
// See www.armresearch.com for the copyright terms.
//
// This is the unit test for the class MailHeaders.
//

#include <iostream>
#include <sstream>
#include <vector>
#include <cstdlib>
#include "MailHeaders.hpp"

using namespace std;

const string SMTPENDL = "\r\n";                                                 // SMTP endline.

/// Output error.
#define Error(msg)                                                         \
    {                                                                      \
        std::cerr << "In file " << __FILE__ << ":" << __LINE__ << ":  ";   \
        std::cerr << msg;                                                  \
    }

/// Exit with error.
#define ErrorExit(msg)                                                     \
    {                                                                      \
        Error(msg)                                                         \
	std::exit(-1);						           \
    }

/// Vector of strings.
typedef vector<string> StringContainer;

/// Test header names
StringContainer TestHeaderName;

/// Test header bodies.
vector<StringContainer> TestHeaderBody;

/// Initialize test input.
//
void
initialize() {
    
    StringContainer HeaderBody;

    HeaderBody.push_back("Body 1 (one line)");                                  // Create the header body.
    TestHeaderName.push_back("X-Header1");                                       // Load the header name...
    TestHeaderBody.push_back(HeaderBody);                                       // ...and body.

    HeaderBody.clear();
    HeaderBody.push_back("Body 2 (line 1 of 2)");
    HeaderBody.push_back("Body 2 (line 2 of 2)");
    TestHeaderName.push_back("X-Header2");
    TestHeaderBody.push_back(HeaderBody);

    HeaderBody.clear();
    HeaderBody.push_back("");
    HeaderBody.push_back("\tBody 3 (line 2 of 2)");
    TestHeaderName.push_back("X-Header3");
    TestHeaderBody.push_back(HeaderBody);

    HeaderBody.clear();
    HeaderBody.push_back("");
    HeaderBody.push_back("");
    HeaderBody.push_back("Body 4 (line 3 of 5)");
    HeaderBody.push_back("");
    HeaderBody.push_back("");
    TestHeaderName.push_back("X-Header4");
    TestHeaderBody.push_back(HeaderBody);
}

/// Create a buffer containing correct headers.
string
createCorrectHeaders() {
    string Headers;                                                             // Contains all the headers.
    string Body;

    StringContainer::iterator iHeader;
    vector<StringContainer>::iterator iBody;

    for (iHeader = TestHeaderName.begin(), iBody = TestHeaderBody.begin();      // For each header...
         iHeader != TestHeaderName.end();
         iHeader++, iBody++) {

        Headers += *iHeader;                                                    // ...add header name...
        Headers += ":";

        Body.clear();
        vector<string>::iterator iLine;
        for (iLine = iBody->begin();
             iLine != iBody->end();
             iLine ++) {

            Body += *iLine + SMTPENDL;                                          // ...construct body...

        }

        Headers += Body;                                                        // ...add body.
    }
    return Headers;
}

/// Test with valid input.
//
// This function tests MailHeaders functionality with correct input.
//
// \returns true if the test passes, false otherwise.
//
bool
testWithCorrectInput() {

    const string TestHeader = createCorrectHeaders();                           // Create correct headers.
    string ExpectedBody;
    string HeaderNameObtained;
    string BodyObtained;

    StringContainer::iterator iHeader;
    vector<StringContainer>::iterator iBody;

    MailHeaders MailHeadersTest;                                                // Object under test.

    MailHeadersTest.loadHeaders(TestHeader);                                    // Load the test headers (to be
                                                                                // parsed) into the object under test.

    for (iHeader = TestHeaderName.begin(), iBody = TestHeaderBody.begin();      // For each header...
         iHeader != TestHeaderName.end();
         iHeader++, iBody++) {

        if (!MailHeadersTest.getNameBody(&HeaderNameObtained, &BodyObtained) || // ...fetch the name/body pair.
            HeaderNameObtained != *iHeader) {                                   // ...compare the name...
            Error("getNameBody failure getting header name.\n");
            return false;
        }

        bool FirstTime;
        FirstTime = true;
        ExpectedBody.clear();                                                   // Assemble the expected body.
        vector<string>::iterator iLine;
        for (iLine = iBody->begin();
             iLine != iBody->end();
             iLine ++) {

            if (FirstTime) {
                ExpectedBody = *iLine;
                FirstTime = false;
            } else {
                ExpectedBody += "\n ";
                ExpectedBody += *iLine;
            }
        }

        if (ExpectedBody != BodyObtained) {
            ostringstream Temp;
            Temp << "getNameBody failure getting header body.  Expected body:  '"
                 << ExpectedBody << "'.  Obtained body:  '" << BodyObtained << "'.\n";
            Error(Temp.str());
            return false;
        }

    }

    if (MailHeadersTest.getNameBody(&HeaderNameObtained, &BodyObtained)) {      // Should be no more headers.
        ostringstream Temp;
        Temp << "getNameBody failure.  Obtained extra header:  '"
             << HeaderNameObtained << "' and body:  '" << BodyObtained << "'.\n";
        Error(Temp.str());
        return false;
    }

    return true;
}

/// Test with invalid input.
//
// This function tests MailHeaders functionality with correct input.
//
// \returns true if the test passes, false otherwise.
//
bool
testWithIncorrectInput() {

    string TestHeader = createCorrectHeaders();                                 // Create correct headers...
    TestHeader += "Bad line";                                                   // ...and add text to cause an error.
    string ExpectedBody;
    string HeaderNameObtained;
    string BodyObtained;

    StringContainer::iterator iHeader;
    vector<StringContainer>::iterator iBody;

    MailHeaders MailHeadersTest;                                                // Object under test.

    MailHeadersTest.loadHeaders(TestHeader);                                    // Load the test headers (to be
                                                                                // parsed) into the object under test.

    try {
        for (iHeader = TestHeaderName.begin(), iBody = TestHeaderBody.begin();  // For each header...
             iHeader != TestHeaderName.end();
             iHeader++, iBody++) {

            if (!MailHeadersTest.getNameBody(&HeaderNameObtained,               // ...fetch the name/body pair.
                                             &BodyObtained) ||
                HeaderNameObtained != *iHeader) {                               // ...compare the name...
                Error("getNameBody failure getting header name.\n");
                return false;
            }

            bool FirstTime;
            FirstTime = true;
            ExpectedBody.clear();                                               // Assemble the expected body.
            vector<string>::iterator iLine;
            for (iLine = iBody->begin();
                 iLine != iBody->end();
                 iLine ++) {

                if (FirstTime) {
                    ExpectedBody = *iLine;
                    FirstTime = false;
                } else {
                    ExpectedBody += "\n ";
                    ExpectedBody += *iLine;
                }
            }

            if (ExpectedBody != BodyObtained) {
                ostringstream Temp;
                Temp << "getNameBody failure getting header body.  Expected body:  '"
                     << ExpectedBody << "'.  Obtained body:  '" << BodyObtained << "'.\n";
                Error(Temp.str());
                return false;
            }

        }
    }
    catch(exception& e) {                                                       // Ignore expected exception.
    }

    try {
        MailHeadersTest.loadHeaders("Bad line #2");                             // Should result in an error.
        (void) MailHeadersTest.getNameBody(&HeaderNameObtained, &BodyObtained);
        Error("Expected exception not thrown by getNameBodyng header body.");
        return false;
    }
    catch(exception& e) {                                                       // Ignore expected exception.
    }

    try {
        MailHeadersTest.loadHeaders(SMTPENDL + SMTPENDL);                       // Should result in an error.
        (void) MailHeadersTest.getNameBody(&HeaderNameObtained, &BodyObtained);
        Error("Expected exception not thrown by getNameBodyng header body.");
        return false;
    }
    catch(exception& e) {                                                       // Ignore expected exception.
    }

    return true;
}

/// Unit tests for MailHeaders.
//
// This function tests the correct operation of the MailHeaders
// functions, and also that they throw exceptions in case of invalid
// input.
//
int main(int argc, char* argv[]) {

    try {                                                                       // Catch anything that breaks loose.

        initialize();                                                           // Initialize test input.

        // Test operation with correct input.
        if (!testWithCorrectInput()) {
            ErrorExit("testWithCorrectInput() failure.\n");
        }

        // Test operation with incorrect input.
        if (!testWithIncorrectInput()) {
            ErrorExit("testWithIncorrectInput() failure.\n");
        }

    }                                                                           // That's all folks.

    catch(exception& e) {                                                       // Report any normal exceptions.
        cerr << "SNFMilter Exception: " << e.what() << endl;
    }
    catch(...) {                                                                // Report any unexpected exceptions.
        cerr << "SNFMilter Panic! Unknown Exception!" << endl;
    }

    return 0;                                                                   // Normally we return zero.
}

